from typing import Optional

from pydantic import BaseModel

from ..utils import Number, StandardUnits, Units


class Coord(BaseModel):
    lat: Number
    """City geo location, latitude"""
    lon: Number
    """City geo location, longitude"""


class Weather(BaseModel):
    id: int
    """Weather condition id. See [here](https://openweathermap.org/weather-conditions#Weather-Condition-Codes-2) for a full list of possibilities."""
    main: str
    """Group of weather parameters (Rain, Snow, Extreme etc.)"""
    description: str
    """Weather condition within the group"""
    icon: str
    """Weather icon id. Icon can be fetched with `http://openweathermap.org/img/wn/{icon}@2x.png`"""


class Main(BaseModel):
    """Do not confuse `temp_min` or `temp_max` with daily lows/highs; see [here](https://openweathermap.org/current#min)"""

    temp: float
    """Temperature"""
    humidity: Number
    """Humidity"""
    feels_like: Number
    temp_min: Number
    """Minimum temperature at the moment. This is deviation from current temp that is possible for large cities and megalopolises geographically expanded (use these parameters optionally)"""
    temp_max: float
    """Maximum temperature at the moment. This is deviation from current temp that is possible for large cities and megalopolises geographically expanded (use these parameters optionally)"""
    pressure: int
    """Atmospheric pressure (on the sea level, if there is no sea_level or grnd_level data)"""
    sea_level: Optional[int] = None
    """Atmospheric pressure on the sea level"""
    grnd_level: Optional[int] = None
    """Atmospheric pressure on the ground level"""


class Wind(BaseModel):
    deg: int
    """Wind direction"""
    speed: float
    """Wind speed. Differs from wind gust in that it measures instantaneous velocity."""
    gust: Optional[float] = None
    """Wind gust. Differs from wind speed in that it measures short bursts in wind."""


class Clouds(BaseModel):
    all: int
    """Cloudiness as a percentage"""


class Sys(BaseModel):
    message: str | None = None
    """System parameter, do not use it"""
    country: Optional[str] = None
    """Country code (GB, JP etc.)"""
    sunrise: int
    """Sunrise time"""
    sunset: int
    """Sunset time"""
    type: Optional[str] = None
    id: Optional[str] = None


class Rain(BaseModel):
    _1h: Number
    _3h: Number


class CurrentWeatherStatus(BaseModel):
    """The weather data served by https://openweathermap.org/current. Most docstrings stolen from there too."""

    units: Units = StandardUnits.STANDARD
    """Units to use. Served by internal API."""
    id: int
    """City identification"""
    dt: int
    """Data receiving time"""
    name: str
    """City name"""
    coord: Coord
    """Geographical coordinates"""
    sys: Sys
    """Location data"""
    wind: Wind
    clouds: Clouds
    weather: list[Weather]
    """A list of weather responses. Usually one item."""
    rain: Optional[Rain] = None
    """Rain data. May not exist."""
    base: str
    main: Main
    visibility: int
    timezone: int
    cod: int



import aiohttp as _aiohttp


Number = int | float


class _AutomaticClient:
    client: _aiohttp.ClientSession
    appid: str | None

    def __init__(self, client: _aiohttp.ClientSession | None = None, appid: str | None = None) -> None:
        self.client = client or _aiohttp.ClientSession()
        self.appid = appid or None

    async def __aenter__(self, *args):
        return self

    async def __aexit__(self, *args):
        await self.client.close()
        return self



from typing import NamedTuple as _NamedTuple
from ._classes import Number


class Units(_NamedTuple):
    temp: tuple[str, str] = ("K", "Kelvin")
    speed: tuple[str, str] = ("m/s", "meter/sec")
    api_name: str = "standard"

    time = ("unix", "UTC")
    pressure = "hPa"
    cloudiness = "%"
    precipitation = ("mm", "millimeters")
    degrees = ("°", "degrees (meteorological)")


class StandardUnits:
    STANDARD = Units()
    METRIC = Units(temp=("°C", "Celsius"), api_name="metric")
    IMPERIAL = Units(temp=("°F", "Fahrenheit"), speed=("mph", "miles/hour"), api_name="imperial")


def convert_temp(temp: Number, __from: Units = StandardUnits.STANDARD, __to: Units = StandardUnits.IMPERIAL) -> Number:
    """Converts temperature between different units"""
    if __from == __to:
        return temp

    match (__from, __to):
        case (StandardUnits.STANDARD, StandardUnits.METRIC):
            return temp - 273.15
        case (StandardUnits.STANDARD, StandardUnits.IMPERIAL):
            return 1.8 * (temp - 273.15) + 32
        case (StandardUnits.METRIC, StandardUnits.STANDARD):
            return temp + 273.15
        case (StandardUnits.METRIC, StandardUnits.IMPERIAL):
            return 1.8 * temp + 32
        case (StandardUnits.IMPERIAL, StandardUnits.STANDARD):
            return (temp - 32) * 1.8 + 273.15
        case (StandardUnits.IMPERIAL, StandardUnits.METRIC):
            return (temp - 32) * 1.8
        case _:
            raise NotImplementedError(
                f"Conversion between types '{__from.__class__}' and '{__to.__class__}' is not defined"
            )


_MPS_PER_MPH = 0.44704


def convert_speed(
    speed: Number, __from: Units = StandardUnits.STANDARD, __to: Units = StandardUnits.IMPERIAL
) -> Number:
    if __from in {StandardUnits.STANDARD, StandardUnits.METRIC} and __to in {
        StandardUnits.STANDARD,
        StandardUnits.METRIC,
    }:
        return speed

    match (__from, __to):
        case (StandardUnits.STANDARD | StandardUnits.METRIC, StandardUnits.IMPERIAL):
            return speed / _MPS_PER_MPH
        case (StandardUnits.IMPERIAL, StandardUnits.STANDARD | StandardUnits.METRIC):
            return _MPS_PER_MPH * speed
        case _:
            raise NotImplementedError(
                f"Conversion between types '{__from.__class__}' and '{__to.__class__}' is not defined"
            )



from ..utils import Number as _Number
from ..utils import Units, _AutomaticClient
from ._classes import *
from .response import *


class CurrentWeatherAPIException(Exception):
    pass


class CurrentWeather(_AutomaticClient):
    BASE_URL = "https://api.openweathermap.org/data/2.5/weather"

    async def get(
        self, coords: tuple[_Number, _Number], units: Units = StandardUnits.STANDARD, lang: str | None = None
    ) -> CurrentWeatherStatus:
        params = {"appid": self.appid, "lat": coords[0], "lon": coords[1], "units": units.api_name}
        if lang:
            params["lang"] = lang

        async with self.client.get(self.BASE_URL, params=params) as resp:
            resp = await resp.json()
            if "cod" in resp and "message" in resp:
                raise CurrentWeatherAPIException(resp["cod"], resp["message"])
            if "rain" in resp:
                keys: set[str] = set(resp["rain"])
                for key in keys:
                    resp["rain"][f"_{key}"] = resp["rain"][key]
                    del resp["rain"][key]

            return CurrentWeatherStatus(**resp, units=units)
